
/*************************************************************************
 *
 * thinkfan version 0.3.3 -- copyleft 10-2008, Victor Mataré
 *
 * A minimalist, lightweight fan control program for modern Thinkpads.
 * So far this is only confirmed to work on my T61p.
 *
 * This program is provided to you AS-IS. You can do with it whatever
 * you want. No warranties whatsoever. If this program steals your
 * car, kills your horse, smokes your dope or pees on your carpet...
 * - too bad, you're on your own.
 *
 *   !!!.. WARNING ..!!!.. WARNING ..!!!.. WARNING ..!!!................
 *   !!!
 *   !!!    This program could damage your hardware in any thinkable
 *   !!!    way. Most likely it will OVERHEAT YOUR CPU.
 *   !!!    There is NO SANITY CHECKING on the configuration.
 *   !!!    That means you can configure it to put your CPU on fire.
 *   !!!    
 *   !!!    Don't run it if you don't understand what it does.
 *
 *
 *   You have been warned.
 *
 *************************************************************************/

#define VERSION "0.3.3"

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>

#define IBM_TEMP "/proc/acpi/ibm/thermal"
#define IBM_FAN "/proc/acpi/ibm/fan"

#define MAXERR 3 // maximum number of errors when trying to set fan speed

typedef struct {
	int level; // this is written into IBM_FAN to set fan level.
	int low;   // lower temperature for stepping down to previous level.
	int high;  // upper temperature determines when to step to the next level.
} thm_tuple;

typedef struct {
	thm_tuple *items;
	int num_items;
} tf_config;

int quiet;
volatile int intr;
unsigned int sleeptime;
unsigned int watchdog_timeout;
char *config_file;

/*******************************************************************
 * get_temp() reads temperatures from /proc/acpi/ibm/thermal and
 * returns the highest one found.
 *******************************************************************/
int get_temp() {
	int i, res, retval=0;
	int *t = calloc(16, sizeof(int));
	FILE *ibm_temp;

	i = 0;
	memset(t, 0, 16);
	if ((ibm_temp = fopen(IBM_TEMP, "r")) == NULL) {
		perror(IBM_TEMP);
		return -1000;
	}
	res = fscanf(ibm_temp,
	"temperatures: %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i %i",
	 &t[i++], &t[i++], &t[i++], &t[i++], &t[i++], &t[i++],
	 &t[i++], &t[i++], &t[i++], &t[i++], &t[i++], &t[i++],
	 &t[i++], &t[i++], &t[i++], &t[i++]);
	fclose(ibm_temp);
	if (res < 2) {
		syslog(LOG_ERR, "Couldn't get two or more temperatures from %s.",
		 IBM_TEMP);
		fprintf(stderr, "Couldn't get two or more temperatures from %s.",
		 IBM_TEMP);
		return -1000;
	}
	for (i=0; i < 16; i++)
		if (t[i] > retval) retval = t[i];
	free(t);
	return retval;
}


/*********************************************************
 * readconfig(char *fname) reads the config file and
 * returns a struct tf_config.
 * Non-matching lines are skipped.
 *********************************************************/
tf_config *readconfig(char* fname) {
	FILE *cfg_file;
	int j, i=0, num=0, level=0, low=0, high=0, errsav;
	tf_config *config;
	config = (tf_config *) malloc(sizeof(tf_config));
	config->items = (thm_tuple *) malloc(sizeof(thm_tuple));
	if ((cfg_file = fopen(fname, "r")) == NULL) {
		errsav = errno;
		perror(fname);
		syslog(LOG_ERR, "%s: %s", fname, strerror(errsav));
		return NULL;
	}
	syslog(LOG_NOTICE, "Config as read from %s:\n", fname);
	syslog(LOG_NOTICE, "Fan level\tLow\tHigh\n");
	if (!quiet) {
		fprintf(stderr, "Config as read from %s:\n", fname);
		fprintf(stderr, "Fan level\tLow\tHigh\n");
	}
	while ((num = fscanf(cfg_file, " ( %d , %d , %d ) ",
	 &config->items[i].level, &config->items[i].low,
	 &config->items[i].high)) != EOF) {
		if (num == 3) {
			syslog(LOG_NOTICE, " %d\t\t%d\t%d\n",
			 config->items[i].level, config->items[i].low,
			 config->items[i].high);
			if (!quiet)
				fprintf(stderr, " %d\t\t%d\t%d\n",
             config->items[i].level, config->items[i].low,
			 config->items[i].high);
			i++;
			if ((config->items = realloc(config->items,
			 (i+1) * sizeof(thm_tuple))) == NULL) {
				syslog(LOG_ERR, "Can't allocate memory!");
				perror("reading config");
				exit(1);
			}
		}
// skip every line that doesn't match the fscanf pattern above
		else while (fgetc(cfg_file) != '\n');
	}
	fclose(cfg_file);
	if (i <= 0) {
		fprintf(stderr, "\nEmpty config or wrong syntax!\n");
		syslog(LOG_ERR, "\nEmpty config or wrong syntax!\n");
		free(config->items);
		free(config);
		return NULL;
	}
	config->num_items = i-1;
	return config;
}

/*********************************************************
 * Checks for fan_control support in thinkpad_acpi and
 * activates the fan watchdog.
 *********************************************************/
int init_fan() {
	char *line = NULL;
	size_t count = 0;
	FILE *ibm_fan;
	
	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		perror(IBM_FAN);
		return 1;
	}
	while (getline(&line, &count, ibm_fan) != -1);
	if (strncmp("commands:", line, 9)) {
		syslog(LOG_ERR, "Module thinkpad_acpi doesn't "
		 "seem to support fan_control\n");
		fprintf(stderr, "\nModule thinkpad_acpi doesn't "
		 "seem to support fan_control\n");
		return -1;
	}
	fprintf(ibm_fan, "watchdog %d\n", watchdog_timeout);
	fclose(ibm_fan);
	free(line);
	return 0;
}

/***********************************************************
 * Set fan speed.
 ***********************************************************/
int setfan(int level) {
	FILE *ibm_fan;

	if ((ibm_fan = fopen(IBM_FAN, "r+")) == NULL) {
		perror(IBM_FAN);
		return 1;
	}
	int ret = fprintf(ibm_fan, "level %d\n", level);
	fclose(ibm_fan);
	if (ret != 8) {
		fprintf(stderr, "Error writing to " IBM_FAN ".\n");
		syslog(LOG_ERR, "Error writing to " IBM_FAN ".\n");
		return 1;
	}
	return 0;
}

/***********************************************************
 * This is the main routine which periodically checks
 * temperatures and adjusts the fan according to config.
 ***********************************************************/
int fancontrol(tf_config *config) {
	int last_temp, temp, level_idx=0, i, errcnt=0, bias=0, diff=0;
	int wt = watchdog_timeout, st = sleeptime, count;

	if (config == NULL) {
		fprintf(stderr, "No configuration!\n");
		syslog(LOG_ERR, "No configuration!\n");
		return 1;
	}
	if (init_fan()) {
		fprintf(stderr, "Error initializing fan watchdog.\n");
		syslog(LOG_ERR, "Error initializing fan watchdog.\n");
		return 1;
	}

// Set initial fan level:
	level_idx = config->num_items;
	temp = get_temp();
	if (temp < 1) {
		fprintf(stderr, "Got a temperature of %d °C. "
		 "This seems dangerous.\n", temp);
		syslog(LOG_ERR, "Got a temperature of %d °C. "
		 "This seems dangerous.", temp);
		return 1;
	}
	while ((temp <= config->items[level_idx].low) \
	 && (level_idx > 0)) level_idx--;
	if (setfan(config->items[level_idx].level)) errcnt++;
	
/**********************************************
 * Main loop. This is the actual fan control.
 **********************************************/
	while(!intr) { // intr is set by the signal handler
		last_temp = temp; // detect temperature jumps
		sleep(st); // st is the variable sleeptime
		temp = get_temp();

// If temperature increased by more than 2 °C since the
// last cycle, we try to react quickly.
		diff = temp - last_temp;
		if (diff >= 2) {
			bias = (st * (diff-1)) / 2;
			if (st > 2) st = 2;
#ifdef DEBUG
			syslog(LOG_DEBUG, "last_temp=%d, temp=%d, biased_temp="
			 "%d, sleeptime=%d", last_temp, temp, temp + bias, st);
			fprintf(stderr, "last_temp=%d, temp=%d, biased_temp="
			 "%d, sleeptime=%d\n", last_temp, temp, temp + bias, st);
#endif
		}
		else if (st < sleeptime) st++;
		temp += bias;
		if (temp >= config->items[level_idx].high) {
			while ((temp >= config->items[level_idx].high) \
			 && (level_idx < config->num_items)) level_idx++;
			if (setfan(config->items[level_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
#ifdef DEBUG
			syslog(LOG_DEBUG, "temp=%d, level=%d, sleeptime=%d",
			 temp, config->items[level_idx].level, st);
			if (!quiet) fprintf(stderr, "temp=%d, level=%d, sleeptime=%d\n",
			 temp, config->items[level_idx].level, st);
#endif
		}
		if (temp <= config->items[level_idx].low) {
			while ((temp <= config->items[level_idx].low) \
			 && (level_idx > 0)) level_idx--;
			if (setfan(config->items[level_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
			st = sleeptime;
#ifdef DEBUG
			syslog(LOG_DEBUG, "temp=%d, level=%d, sleeptime=%d",
			 temp, config->items[level_idx].level, st);
			if (!quiet) fprintf(stderr, "temp=%d, level=%d, sleeptime=%d\n",
			 temp, config->items[level_idx].level, st);
#endif
		}

// slowly reduce the bias
		if (bias > 0) bias--;

// Write current fan level to IBM_FAN one cycle before watchdog timeout
// ends, to let it know we're alive.
		if ( (wt -= st) <= st ) {
			if (setfan(config->items[level_idx].level))
				if (++errcnt >= MAXERR) return errcnt;
			wt = watchdog_timeout;
		}

	}
	return 0;
}

void sigHandler(int signum) {
	switch(signum) {
	case SIGHUP:
		intr = signum;
		break;
	case SIGINT:
	case SIGTERM:
		intr = signum;
		break;
	}
}

void usage() {
	fprintf(stderr, "\nthinkfan %s - A minimalist "
	 "Thinkpad fan controller\n", VERSION);
	fprintf(stderr, "\nUsage: thinkfan [-h | [-c C"
	"ONFIG] [-s SECONDS]]\n");
	fprintf(stderr, "\n -h\tThis help message");
	fprintf(stderr, "\n -s\tMaximum time between "
	 "temperature reads (default: 5)");
	fprintf(stderr, "\n -c\tLoad configuration from "
	 "file CONFIG (default: /etc/thinkfan.conf)");
	fprintf(stderr, "\n -n\tDo not fork() into the "
	 "background");
	fprintf(stderr, "\n -q\tBe quiet (no status info"
	 " on terminal)\n\n");
}


/***************************************************************
 * Main function:
 * Scan for arguments, set options and initialize signal handler
 ***************************************************************/
int main(int argc, char **argv) {
	int opt, ret, nodaemon=0;
	char *invalid = "";
	sleeptime = 5;
	quiet = 0;
	intr = 0;
	struct sigaction handler;

	config_file = "/etc/thinkfan.conf";
	openlog("thinkfan", LOG_CONS, LOG_USER);
	syslog(LOG_INFO, "thinkfan " VERSION " starting...");

	while ((opt = getopt(argc, argv, "c:s:hnq")) != -1) {
		switch(opt) {
		case 'h':
			usage();
			return 0;
			break;
		case 'n':
			nodaemon = 1;
			break;
		case 'c':
			config_file = optarg;
			break;
		case 's':
			sleeptime = (unsigned int) strtol(optarg, &invalid, 0);
			if (*invalid != 0) {
				fprintf(stderr, "\nERROR: option -s requires numerical argument!\n");
				usage();
				return 1;
			}
			break;
		case 'q':
			quiet = 1;
			break;
		default:
			usage();
			return 1;
		}
	}

	memset(&handler, 0, sizeof(handler));
	handler.sa_handler = sigHandler;
	if (sigaction(SIGHUP, &handler, NULL) \
	 || sigaction(SIGINT, &handler, NULL) \
	 || sigaction(SIGTERM, &handler, NULL)) perror("sigaction");
	watchdog_timeout = sleeptime * 6;

	if (nodaemon) return run();
	else {
		if (!(ret = fork())) return run();
		else if (ret == -1) {
			perror("fork()");
			return 1;
		}
	}
	return 0;
}


/********************************************************************
 * Outer loop. Handles signal conditions, runtime cleanup and config
 * reloading
 ********************************************************************/
int run() {
	int ret = 0;
	tf_config *config, *newconfig=NULL;
	FILE *pidfile;

	if ((config = readconfig(config_file)) == NULL) {
		fprintf(stderr, "Refusing to run without config file. Please read the documentation.\n");
		syslog(LOG_ERR, "Refusing to run without config file. Please read the documentation.");
		return -2;
	}

	if ((pidfile = fopen("/var/run/thinkfan.pid", "w+")) == NULL) {
		perror("Opening /var/run/thinkfan.pid");
		return -1;
	}
	fprintf(pidfile, "%d", getpid());
	fclose(pidfile);

	while ((intr < SIGINT) && !ret) {
		intr = 0;
		ret = fancontrol(config);
		if (ret) {
			fprintf(stderr, "\nCounted %d error(s) during fancontrol."
			 " Exiting.\n", ret);
			syslog(LOG_ERR, "Counted %d error(s) during fancontrol."
			 " Exiting.", ret);
			break;
		}
		else if (intr < SIGINT) {
			syslog(LOG_DEBUG, "Reloading config...");
			if (!quiet)
				fprintf(stderr, "\nReloading config...\n");
			if ((newconfig = readconfig(config_file)) != NULL) {
				free(config->items);
				free(config);
				config = newconfig;
			}
			else {
				fprintf(stderr, "\nError reloading config. Keeping old one.\n");
				syslog(LOG_ERR, "Error reloading config. Keeping old one.");
			}
		}
		else {
			if (!quiet)
				fprintf(stderr, "\nCaught deadly signal. Cleaning up"
				 " and resetting fan control.\n");
			syslog(LOG_WARNING, "Caught deadly signal. Cleaning up"
			 " and resetting fan control.");
			break;
		}
	}
	free(config->items);
	free(config);
	FILE *fan;
	if ((fan = fopen(IBM_FAN, "r+")) == NULL) {
		int errsav = errno;
		perror("Opening " IBM_FAN);
		syslog(LOG_ERR, "Opening " IBM_FAN ": %s", strerror(errsav));
		return -3;
	}
	fprintf(fan, "level auto\n");
	fclose(fan);
	remove("/var/run/thinkfan.pid");
	return ret;
}

